/*
 * SourceMod Radio
 *
 * Adds a radio player to SourceMod that supports direct station links or streaming audio
 *
 * Coded by dubbeh - www.dubbeh.net
 *
 * Licensed under the GPLv3
 *
 */

#pragma semicolon 1

#include <sourcemod>
#include <clientprefs>
#include <multicolors>

#pragma newdecls required

#undef REQUIRE_PLUGIN
#undef REQUIRE_EXTENSIONS
#include <updater>
#include <SteamWorks>
#include <motdf>
#define REQUIRE_EXTENSIONS
#define REQUIRE_PLUGIN

#define PLUGIN_NAME					"Radio"
#define PLUGIN_VERSION				"2.0.0.15"
#define MAX_MOTD_URL_SIZE			192 // Maximum MOTD message size is 192 bytes

// Radio play types
#define RADIO_PLAY_TYPE_DIRECT 			1
#define RADIO_PLAY_TYPE_VOLUME_CONTROL 	2
#define RADIO_PLAY_TYPE_SHOUTCAST		3

// Check for playlist updates every 30 minutes - 1800 seconds
#define SHOUTCAST_UPDATE_CHECK_TIME 1800.0
#define TRACK_UPDATE_TIME 			30.0

#define SHOUTCAST_PLAYLIST_TYPE_STANDARD 	1
#define SHOUTCAST_PLAYLIST_TYPE_NO_CHILDREN 2
#define SHOUTCAST_PLAYLIST_TYPE_TOP_500		3

// Thanks to GoD-Tony for this macro
#define STEAMWORKS_AVAILABLE()	(GetFeatureStatus(FeatureType_Native, "SteamWorks_IsLoaded") == FeatureStatus_Available)

public Plugin myinfo = 
{
	name = PLUGIN_NAME,
	author = "dubbeh",
	description = "Radio stations plugin for SourceMod - Powered by ShoutCast.com",
	version = PLUGIN_VERSION,
	url = "www.dubbeh.net"
};

// Storage Method-map includes
#include "radio/station_storage.inc"
#include "radio/genre_storage.inc"

/* cVar Handles */
ConVar g_cVarEnable = null;
ConVar g_cVarStationAdvert = null;
ConVar g_cVarWelcomeMsg = null;
ConVar g_cVarLogging = null;
ConVar g_cVarAutoplay = null;
ConVar g_cVarInitialConnectVolume = null;
ConVar g_cVarAutoUpdate = null;
ConVar g_cVarCheckForHTMLMOTD = null;
ConVar g_cVarPlayType = null;
ConVar g_cVarShoutCastCustom = null;
ConVar g_cVarEnableTrackQuery = null;
ConVar g_cVarShoutCastPlayListType = null;
ConVar g_cVarVolumeWrapperType = null;
ConVar g_cVarNoMissingMOTDFMsgs = null;

/* Static menu handles */
Menu g_mRadioBaseMenu = null;
Menu g_mRadioBaseGenresMenu = null;
Menu g_mRadioCustomBaseGenresMenu = null;
Menu g_mRadioStationsMenu = null;
Menu g_mRadioCustomStationsMenu = null;
Menu g_mRadioVolumeMenu = null;

/* Configs and playlist type file names */
char g_szRadioStationsDirectBaseFile[] = "radiodirect.txt";
char g_szRadioStationsVolumeBaseFile[] = "radiovolume.txt";
char g_szRadioStationsShoutCastBaseFile[] = "radioshoutcast.txt";
char g_szRadioStationsShoutCastCustomFile[] = "shoutcastcustom.txt";
char g_szConfigFile[] = "sourcemod/plugin.radio.cfg";

/* URL's for various settings */
char g_szRadioOffPage[MAX_MOTD_URL_SIZE + 1] = "https://radio.dubbeh.net/radiooff.html";
char g_szVolumeWrappersBaseURL[MAX_MOTD_URL_SIZE + 1] = "https://radio.dubbeh.net";
char g_szUpdateURL[] = "https://update.dubbeh.net/radio2/radio.txt";
char g_szShoutCastURL[] = "https://radio.dubbeh.net/index.php";
char g_szShoutCastQueryURL[] = "https://radio.dubbeh.net/query.php";
char g_szShoutCastVersionURL[] = "https://radio.dubbeh.net/stationversion.txt";
char g_szShoutCastPlaylistStandardURL[] = "https://radio.dubbeh.net/radiostations.txt";
char g_szShoutCastPlaylistTop500URL[] = "https://radio.dubbeh.net/top500.txt";
char g_szShoutCastPlaylistNoChildrenURL[] = "https://radio.dubbeh.net/nochildren.txt";
char g_szVolumeControlTrackQueryURL[MAX_MOTD_URL_SIZE + 1] = "https://radio.dubbeh.net/trackquery.php";

/* Strings that are printed multiple times */
char g_szInvalidStationIndex[] = "[SM-RADIO] Error: Tried to load a station with an invalid index. Please check your stations list file.";
char g_szNoStreamURLSet[] = "[SM-RADIO] Error: Station with ID \"%d\" has no Stream URL set. Please correct this in the config file.";

/* Timer handles */
Handle g_hUpdateTimer = INVALID_HANDLE;

/* Everything else */
int g_iShoutCastPlaylistVersion = 0;
EngineVersion g_EngineVersion = Engine_Unknown;
bool g_bUpdaterAvail = false;
bool g_bMOTDFAvail = false;
bool g_bUseGenres = true;
bool g_bCustUseGenres = true;
bool g_bWriteKVToDisk = false;
bool g_bStationsParsing = false;
bool g_bPlayListFound = false;
bool g_bDisabledHTMLMOTD[MAXPLAYERS + 1];

// Cached cVars that can't be changed during maps - This is more of a safety precaution
int g_iPlayType = RADIO_PLAY_TYPE_SHOUTCAST;
bool g_bShoutCastCustom = false;

// Storage Method-maps
Stations g_Stations;
Genres g_Genres;
CustomStations g_CustStations;
CustomGenres g_CustGenres;

#include "radio/client_cookies.inc"
ClientCookies g_Cookies;

#include "radio/helpers.inc"
#include "radio/menu_handlers.inc"
#include "radio/menus_create.inc"
#include "radio/commands.inc"
#include "radio/file.inc"
#include "radio/timers.inc"
#include "radio/track_query.inc"
#include "radio/updater.inc"

public void OnPluginStart()
{
	CreateConVar("sm_radio_version", PLUGIN_VERSION, "SourceMod Radio version", FCVAR_NOTIFY | FCVAR_DONTRECORD);
	g_cVarEnable = CreateConVar("sm_radio_enable", "1.0", "Enable SourceMod Radio", 0, true, 0.0, true, 1.0);
	g_cVarStationAdvert = CreateConVar("sm_radio_advert", "1.0", "Enable advertising the users radio station choice", 0, true, 0.0, true, 1.0);
	g_cVarWelcomeMsg = CreateConVar("sm_radio_welcome", "1.0", "Enable the welcome message", 0, true, 0.0, true, 1.0);
	g_cVarLogging = CreateConVar("sm_radio_logging", "1.0", "Enable logging to the server console and log file", 0, true, 0.0, true, 1.0);
	g_cVarAutoplay = CreateConVar("sm_radio_autoplay", "1.0", "Enable auto playing a users radio choice on map change", 0, true, 0.0, true, 1.0);
	g_cVarInitialConnectVolume = CreateConVar("sm_radio_init_conn_vol", "0.10", "Initial volume set for a user on first connect", 0, true, 0.01, true, 1.0);
	g_cVarAutoUpdate = CreateConVar("sm_radio_auto_update", "1.0", "Enable the plugin to automatically update (if available updates found)", 0, true, 0.0, true, 1.0);
	g_cVarCheckForHTMLMOTD = CreateConVar("sm_radio_check_html_motd", "1.0", "Enable to check if \"cl_disablehtmlmotd\" is enabled and warn the user", 0, true, 0.0, true, 1.0);
	g_cVarPlayType = CreateConVar("sm_radio_play_type", "3.0", "Play type for Radio, 1 = Direct, 2 = Volume Wrapper, 3 = ShoutCast Powered", 0, true, 1.0, true, 3.0);
	g_cVarShoutCastCustom = CreateConVar("sm_radio_shoutcast_custom", "0.0", "Append custom stations using the wrapper page under ShoutCast mode from shoutcastcustom.txt", 0, true, 0.0, true, 1.0);
	g_cVarEnableTrackQuery = CreateConVar("sm_radio_track_query_enable", "1.0", "Enable the track query feature", 0, true, 0.0, true, 1.0);
	g_cVarShoutCastPlayListType = CreateConVar("sm_radio_shoutcast_playlist", "3.0", "ShoutCast playlist type to utilise (1 = Standard, 2 = No Children, 3 = Top 500)", 0, true, 1.0, true, 3.0);
	g_cVarVolumeWrapperType = CreateConVar("sm_radio_volume_wrapper_type", "0.0", "Radio volume wrapper type to use. 0 is Standard <audio> HTML5 - 1 is JPlayer", 0, true, 0.0, true, 1.0);
	g_cVarNoMissingMOTDFMsgs = CreateConVar("sm_no_missing_motdf_msgs", "0.0", "Stop displaying the missing MOTD Fixer messages when trying to load a station under certain mods", 0, true, 0.0, true, 1.0);
	
	LoadTranslations("radio.phrases");
	
	RegConsoleCmd("sm_radio", Command_Radio);
	RegConsoleCmd("sm_radiooff", Command_RadioOff);
	RegConsoleCmd("sm_roff", Command_RadioOff);
	RegConsoleCmd("sm_browse", Command_Browse);
	RegConsoleCmd("sm_volume", Command_Volume);
	RegConsoleCmd("sm_vol", Command_Volume);
	RegConsoleCmd("sm_track", Command_Track);
	RegConsoleCmd("sm_song", Command_Track);
	RegConsoleCmd("sm_genres", Command_Genres);
	RegConsoleCmd("sm_genre", Command_Genres);
	RegConsoleCmd("sm_radiohelp", Command_Help);
	RegConsoleCmd("sm_rhelp", Command_Help);

	RegAdminCmd("sm_radio_force_download", Command_ForcePlaylistDownload, ADMFLAG_RCON, "Force download of the playlist type");
	
	for (int iIndex = 0; iIndex <= MAXPLAYERS; iIndex++) {
		g_bDisabledHTMLMOTD[iIndex] = false;
	}

	g_Cookies.Init();
	g_EngineVersion = GetEngineVersion();
	
	if (g_EngineVersion == Engine_CSGO)
	{
		RLogMessage("[SM-RADIO] Running under CS:GO detected");
	}
	else if (g_EngineVersion == Engine_Unknown)
	{
		RLogMessage("[SM-RADIO] Unknown mod found. Please contact the SourceMod authors for support.");
		SetFailState("[SM-RADIO] Unknown mod found. Please contact the SourceMod authors for support.");
	}
	
	// Auto create the config file, if it doesn't exist
	AutoExecConfig(true, "plugin.radio", "sourcemod");
	
	ServerCommand("exec %s", g_szConfigFile);
	
	// Create the radioshoutcast.txt auto-updater schedule - This repeats every 30 minutes
	g_hUpdateTimer = CreateTimer(SHOUTCAST_UPDATE_CHECK_TIME, ShoutcastUpdateListTimer, INVALID_HANDLE, TIMER_REPEAT);
	
	if (g_hUpdateTimer == INVALID_HANDLE)
		SetFailState("[SM-RADIO] Error: Unable to create the ShoutCast update list timer.");
}

public void OnAllPluginsLoaded()
{
	g_bUpdaterAvail = LibraryExists("updater");
	
	if (g_bUpdaterAvail)
		RLogMessage("[SM-RADIO] Updater plugin found.");
		
	g_bMOTDFAvail = LibraryExists("motdf");
	
	if (g_bMOTDFAvail)
		RLogMessage("[SM-RADIO] MOTD Fixer plugin found.");

	if (STEAMWORKS_AVAILABLE())
		RLogMessage("[SM-RADIO] SteamWorks extension found.");
	else
		RLogMessage("[SM-RADIO] Please install SteamWorks to keep ShoutCast playlists in sync.");

	if (g_EngineVersion == Engine_CSGO && !g_bMOTDFAvail && !g_cVarNoMissingMOTDFMsgs.BoolValue)
		RLogMessage("[SM-RADIO] Please install MOTD Fixer to restore CS:GO functionality - Download from: https://dubbeh.net/sourcemod-plugins/motd-fixer/");
}

public void OnLibraryAdded(const char[] name)
{
	if (LibraryExists("updater"))
		g_bUpdaterAvail = true;
		
	if (LibraryExists("motdf"))
		g_bMOTDFAvail = true;
}

public void OnLibraryRemoved(const char[] name)
{
	if (StrEqual(name, "updater"))
		g_bUpdaterAvail = false;
	
	if (StrEqual(name, "motdf"))
		g_bMOTDFAvail = false;
}

public void OnPluginEnd()
{
	if (g_hUpdateTimer != INVALID_HANDLE)
		CloseHandle(g_hUpdateTimer);
}

public void OnMapStart()
{
	// Make sure the config file is executed
	ServerCommand("exec %s", g_szConfigFile);
}

// Moved from OnMapStart to OnConfigsExecuted for plugin intialization (Much more reliable here)
public void OnConfigsExecuted()
{
	// Cache the cVars on map start to avoid problems
	g_iPlayType = g_cVarPlayType.IntValue;
	g_bShoutCastCustom = g_cVarShoutCastCustom.BoolValue;
	
	// Map Init variables
	g_bUseGenres = false;
	g_bCustUseGenres = false;
	g_bPlayListFound = false;
	
	// Load the stations key values file(s)
	GetRadioStationsFromFile(false);
	
	// Load the custom ShoutCast stations file
	if (g_bShoutCastCustom && g_iPlayType == RADIO_PLAY_TYPE_SHOUTCAST && g_bPlayListFound)
	    GetRadioStationsFromFile(true);
	
	// Create the radio menus
	CreateRadioMenus();
	
	if (g_cVarAutoUpdate.BoolValue)
	{
		if (g_bUpdaterAvail)
		{
			Updater_AddPlugin(g_szUpdateURL);
		}
		else
		{
			RLogMessage("[SM-RADIO] Auto-Updating enabled but unable to find the SourceMod Updater plugin.");
			RLogMessage("[SM-RADIO] Download it from : https://forums.alliedmods.net/showthread.php?t=169095");
		}
	}
}

public void OnMapEnd()
{
	if (g_mRadioBaseMenu) { delete g_mRadioBaseMenu; }
	if (g_mRadioBaseGenresMenu) { delete g_mRadioBaseGenresMenu; }
	if (g_mRadioCustomBaseGenresMenu) { delete g_mRadioCustomBaseGenresMenu; }
	if (g_mRadioStationsMenu) { delete g_mRadioStationsMenu; }
	if (g_mRadioCustomStationsMenu) { delete g_mRadioCustomStationsMenu; }
	if (g_mRadioVolumeMenu) { delete g_mRadioVolumeMenu; }
	
	if (g_bPlayListFound)
	{
		g_Stations.Clear();
		if (g_bUseGenres)
		    g_Genres.Clear();
		
		if (g_bShoutCastCustom && g_iPlayType == RADIO_PLAY_TYPE_SHOUTCAST)
		{
			g_CustStations.Clear();
			if (g_bCustUseGenres)
			    g_CustGenres.Clear();
		}
	}
}

public void OnClientPostAdminCheck(int iClient)
{
	g_bDisabledHTMLMOTD[iClient] = false;
	
	if ((!iClient) || !g_cVarEnable.BoolValue || !IsClientConnected(iClient) || IsFakeClient(iClient))
		return;
	
	// Extra check here - Just incase we don't need to waste a thread
	if (g_cVarWelcomeMsg.BoolValue || g_cVarAutoplay.BoolValue || g_cVarCheckForHTMLMOTD.BoolValue)
		CreateTimer(30.0, WelcomeAdvertTimer, GetClientSerial(iClient));
}

// Thanks to psychonic for the cl_disablehtmlmotd query code - modified it a little
public int CheckDisabledHTMLMOTD(QueryCookie qcCookie, int iClient, ConVarQueryResult cqrResult, const char[] szCvarName, const char[] szCvarValue)
{
	if (cqrResult == ConVarQuery_Okay && StringToInt(szCvarValue) == 1)
	{
		g_bDisabledHTMLMOTD[iClient] = true;
		CPrintToChat(iClient, "%T", "Disabled HTML MOTD On", LANG_SERVER);
	}
}

/* Thanks to GoD-Tony for this - Modified for SteamWorks */
public APLRes AskPluginLoad2(Handle myself, bool late, char[] error, int err_max)
{
	// SteamWorks Natives
	MarkNativeAsOptional("SteamWorks_IsLoaded");
	MarkNativeAsOptional("SteamWorks_CreateHTTPRequest");
	MarkNativeAsOptional("SteamWorks_GetHTTPResponseBodyCallback");
	MarkNativeAsOptional("SteamWorks_SetHTTPCallbacks");
	MarkNativeAsOptional("SteamWorks_SetHTTPRequestHeaderValue");
	MarkNativeAsOptional("SteamWorks_SendHTTPRequest");
	MarkNativeAsOptional("SteamWorks_WriteHTTPResponseBody");
	MarkNativeAsOptional("SteamWorks_ReleaseHTTPRequest");
	return APLRes_Success;
}

